@@ -10,6 +10,7 @@ const Main = imports.ui.main;
 const Applet = imports.ui.applet;
 const Extension = imports.ui.extension;
 const ModalDialog = imports.ui.modalDialog;
+const Gettext = imports.gettext;
 
 // Maps uuid -> metadata object
 var appletMeta;
@@ -29,7 +30,7 @@ var appletsLoaded = false;
 const Roles = {
     NOTIFICATIONS: 'notifications',
     PANEL_LAUNCHER: 'panellauncher'
-}
+};
 
 let enabledAppletDefinitions;
 let clipboard = [];
@@ -39,14 +40,14 @@ function init() {
     appletMeta = Extension.Type.APPLET.maps.meta;
 
     appletsLoaded = false;
-    
+
     // Load all applet extensions, the applets themselves will be added in finishExtensionLoad
     enabledAppletDefinitions = getEnabledAppletDefinitions();
     for (let uuid in enabledAppletDefinitions.uuidMap) {
         Extension.loadExtension(uuid, Extension.Type.APPLET);
     }
     appletsLoaded = true;
-    
+
     global.settings.connect('changed::enabled-applets', onEnabledAppletsChanged);
 }
 
@@ -80,10 +81,10 @@ function getEnabledAppletDefinitions() {
         // maps applet_id -> single applet definition
         idMap: []
     };
-    
+
     // Upgrade settings if required
     checkForUpgrade(result.raw);
-    
+
     // Parse all definitions
     for (let i=0; i<result.raw.length; i++) {
         let appletDefinition = getAppletDefinition(result.raw[i]);
@@ -94,7 +95,7 @@ function getEnabledAppletDefinitions() {
             result.idMap[appletDefinition.applet_id] = appletDefinition;
         }
     }
-    
+
     return result;
 }
 
@@ -118,8 +119,8 @@ function getAppletDefinition(definition) {
             orientation = setOrientationForPanel(panel.panelPosition);
             location = getLocation(panel, elements[1]);
         }
-        
-        return {
+
+        let appletDefinition = {
             panel: panel,
             panelId: panelId,
             orientation: orientation,
@@ -130,6 +131,9 @@ function getAppletDefinition(definition) {
             uuid: elements[3],
             applet_id: elements[4]
         };
+
+        if (elements.length > 5) appletDefinition.overrides = elements[5].split(',');
+        return appletDefinition;
     }
 
     global.logError("Bad applet definition: " + definition);
@@ -204,23 +208,23 @@ function onEnabledAppletsChanged() {
         // Remove all applet instances that do not exist in the definition anymore.
         for (let applet_id in oldEnabledAppletDefinitions.idMap) {
             if(!enabledAppletDefinitions.idMap[applet_id]) {
-                let removeConfig = Extension.get_max_instances(oldEnabledAppletDefinitions.idMap[applet_id].uuid, Extension.Type.APPLET) != 1
+                let removeConfig = Extension.get_max_instances(oldEnabledAppletDefinitions.idMap[applet_id].uuid, Extension.Type.APPLET) != 1;
                 removeAppletFromPanels(oldEnabledAppletDefinitions.idMap[applet_id], removeConfig);
             }
         }
-        
+
         // Unload all applet extensions that do not exist in the definition anymore.
         for (let uuid in oldEnabledAppletDefinitions.uuidMap) {
             if(!enabledAppletDefinitions.uuidMap[uuid]) {
                 Extension.unloadExtension(uuid, Extension.Type.APPLET);
             }
         }
-        
+
         // Add or move applet instances of already loaded applet extensions
         for (let applet_id in enabledAppletDefinitions.idMap) {
             let newDef = enabledAppletDefinitions.idMap[applet_id];
             let oldDef = oldEnabledAppletDefinitions.idMap[applet_id];
-            
+
             if(!oldDef || !appletDefinitionsEqual(newDef, oldDef)) {
                 let extension = Extension.Type.APPLET.maps.objects[newDef.uuid];
                 if(extension) {
@@ -228,7 +232,7 @@ function onEnabledAppletsChanged() {
                 }
             }
         }
-        
+
         // Make sure all applet extensions are loaded.
         // Once loaded, the applets will add themselves via finishExtensionLoad
         for (let uuid in enabledAppletDefinitions.uuidMap) {
@@ -246,7 +250,7 @@ function removeAppletFromPanels(appletDefinition, deleteConfig) {
     let applet = appletObj[appletDefinition.applet_id];
     if (applet) {
         try {
-            applet._onAppletRemovedFromPanel();
+            applet._onAppletRemovedFromPanel(deleteConfig);
         } catch (e) {
             global.logError("Error during on_applet_removed_from_panel() call on applet: " + appletDefinition.uuid + "/" + appletDefinition.applet_id, e);
         }
@@ -256,7 +260,10 @@ function removeAppletFromPanels(appletDefinition, deleteConfig) {
             applet._panelLocation = null;
         }
 
-        delete applet._extension._loadedDefinitions[appletDefinition.applet_id];
+        if (applet._extension) {
+           delete applet._extension._loadedDefinitions[appletDefinition.applet_id];
+        }
+
         delete appletObj[appletDefinition.applet_id];
 
         if (deleteConfig)
@@ -279,7 +286,7 @@ function _removeAppletConfigFile(uuid, instanceId) {
     let file = Gio.File.new_for_path(config_path);
     if (file.query_exists(null)) {
         try {
-            file.delete(null, null);
+            file.delete(null);
         } catch (e) {
             global.logError("Problem removing applet config file during cleanup.  UUID is " + uuid + " and filename is " + config_path);
         }
@@ -301,22 +308,32 @@ function addAppletToPanels(extension, appletDefinition) {
         applet._order = appletDefinition.order;
         applet._extension = extension;
 
+        let location = appletDefinition.location;
+
         // Remove it from its previous panel location (if it had one)
-        if (applet._panelLocation != null) {
+        if (applet._panelLocation != null && location != applet._panelLocation) {
             applet._panelLocation.remove_actor(applet.actor);
             applet._panelLocation = null;
         }
 
-        let location = appletDefinition.location;
-
         let before = location.get_children()
-            .find(x => (x._applet instanceof Applet.Applet) &&
-                       (appletDefinition.order < x._applet._order));
+            .find(x => {
+                return x._applet && (x._applet instanceof Applet.Applet) && (appletDefinition.order < x._applet._order)
+            });
 
-        if (before)
-            location.insert_child_below(applet.actor, before);
-        else
-            location.add_actor(applet.actor);
+        if (before) {
+            if (applet._panelLocation == null) {
+                location.insert_child_below(applet.actor, before);
+            } else {
+                location.set_child_below_sibling(applet.actor, before);
+            }
+        } else {
+            if (applet._panelLocation == null) {
+                location.add_actor(applet.actor);
+            } else {
+                location.set_child_above_sibling(applet.actor, null);
+            }
+        }
 
         applet._panelLocation = location;
 
@@ -340,34 +357,125 @@ function addAppletToPanels(extension, appletDefinition) {
 function removeAppletFromInappropriatePanel (extension, applet, appletDefinition) {
     //  We want to ensure that applets placed in a panel can be shown correctly
     //  - particularly because wide applets will not fit in a vertical panel unless
-    //  they have logic to manage this explicitly
-    //  If the applet is of type Icon Applet (and not a text icon applet) then should be fine otherwise
-    //  we look to see if it has declared itself suitable via a getAllowedLayout call
+    //  they have logic to manage this explicitly.
     //
-    //  If the applet turns out to be unsuitable then remove it.  The applet will show with a red
-    //  indicator in the applet list
-    if (applet instanceof Applet.IconApplet && !(applet instanceof Applet.TextIconApplet)) {
-        ;
-    } else {
-        let allowedLayout = applet.getAllowedLayout();
+    //  If the applet is of type IconApplet (and not a TextIconApplet) then it should be fine.
+    //  If not, we check if the user has previously opted to leave it there anyway.
+    //  Then we look to see if it has declared itself suitable via a call to applet.getAllowedLayout().
+    //
+    //  If the applet turns out to be unsuitable the user is then asked if they want to keep it anyway,
+    //  remove it, or try to find another panel that supports it.
+    if (applet instanceof Applet.IconApplet && !(applet instanceof Applet.TextIconApplet)) return;
+    if (appletDefinition.overrides && appletDefinition.overrides.indexOf('orient') != -1) return;
 
-        if ((appletDefinition.orientation == St.Side.LEFT || appletDefinition.orientation == St.Side.RIGHT) &&
-             allowedLayout == Applet.AllowedLayout.HORIZONTAL) {
-            global.logError("applet " + appletDefinition.uuid + " not suitable for panel orientation " + appletDefinition.orientation);
-            removeAppletFromPanels(extension._loadedDefinitions[appletDefinition.applet_id]);
-            let dialog = new ModalDialog.NotifyDialog(_("This applet is not suitable for vertical panels") + "\n\n");
-            dialog.open();
-        } else if ((appletDefinition.orientation == St.Side.TOP || appletDefinition.orientation == St.Side.BOTTOM) &&
-                    allowedLayout == Applet.AllowedLayout.VERTICAL) {
-                global.logError("applet " + appletDefinition.uuid + " not suitable for panel orientation " + appletDefinition.orientation);
-                removeAppletFromPanels(extension._loadedDefinitions[appletDefinition.applet_id]);
-                let dialog = new ModalDialog.NotifyDialog(_("This applet is not suitable for horizontal panels") + "\n\n");
-                dialog.open();
-        }
+    let allowedLayout = applet.getAllowedLayout();
+
+    if ((allowedLayout == Applet.AllowedLayout.HORIZONTAL && [St.Side.LEFT, St.Side.RIGHT].indexOf(appletDefinition.orientation) != -1) ||
+        (allowedLayout == Applet.AllowedLayout.VERTICAL && [St.Side.TOP, St.Side.BOTTOM].indexOf(appletDefinition.orientation) != -1)) {
+
+        global.logWarning((allowedLayout == Applet.AllowedLayout.HORIZONTAL)+", "+[St.Side.LEFT, St.Side.RIGHT].indexOf(appletDefinition.orientation));
+
+        let label_text = "<b>" + extension.meta.name + "</b>\n" +
+                         _("This applet does not support panels of that type. This can cause visual glitches in the panel.") + "\n" +
+                         _("Would you like to continue using it anyway, remove it from the panel, or try to move it to a different panel?");
+        let label = new St.Label({text: label_text});
+        label.clutter_text.set_use_markup(true);
+
+        let dialog = new ModalDialog.ModalDialog();
+        dialog.contentLayout.add(label);
+
+        dialog.setButtons([
+            {
+                label: _("Leave it"),
+                action: function() {
+                    dialog.destroy();
+                    verticalPanelOverride(appletDefinition);
+                }
+            },
+            {
+                label: _("Remove it"),
+                action: function() {
+                    dialog.destroy();
+                    removeApplet(appletDefinition);
+                }
+            },
+            {
+                label: _("Move to another panel"),
+                action: function() {
+                    dialog.destroy();
+                    moveApplet(appletDefinition, allowedLayout);
+                }
+            }
+        ]);
+
+        dialog.open();
     }
 }
 
+function verticalPanelOverride(appletDefinition) {
+    let list = global.settings.get_strv('enabled-applets');
+    for (let i = 0; i < list.length; i++) {
+        let info = list[i].split(':');
+        if (info[3] == appletDefinition.uuid && info[4] == appletDefinition.applet_id) {
+            let overrides;
+            if (info.length > 5) overrides = info[5].split;
+            else overrides = [];
+            overrides.push('orient');
+            info[5] = overrides.join(',');
+            list[i] = info.join(':');
+            break;
+        }
+    }
 
+    global.settings.set_strv('enabled-applets', list);
+}
+
+function removeApplet(appletDefinition) {
+    let oldList = global.settings.get_strv('enabled-applets');
+    let newList = []
+
+    for (let i = 0; i < oldList.length; i++) {
+        let info = oldList[i].split(':');
+        if (info[3] != appletDefinition.uuid || info[4] != appletDefinition.applet_id) {
+            newList.push(oldList[i]);
+        }
+    }
+
+    global.settings.set_strv('enabled-applets', newList);
+}
+
+function moveApplet(appletDefinition, allowedLayout) {
+    let panelId = null;
+    let panels = global.settings.get_strv('panels-enabled');
+    for (let i = 0; i < panels.length; i++) {
+        let panelInfo = panels[i].split(':');
+        global.logWarning(allowedLayout==Applet.AllowedLayout.HORIZONTAL);
+        if ((allowedLayout == Applet.AllowedLayout.HORIZONTAL && ['top', 'bottom'].indexOf(panelInfo[2]) != -1) ||
+            (allowedLayout == Applet.AllowedLayout.VERTICAL && ['left', 'right'].indexOf(panelInfo[2]) != -1)) {
+            panelId = 'panel' + panelInfo[0];
+            break;
+        }
+    }
+
+    if (panelId == null) {
+        removeApplet(appletDefinition);
+        let dialog = new ModalDialog.NotifyDialog(_("A suitable panel could not be found. The applet has been removed instead.") + "\n\n");
+        dialog.open();
+        return;
+    }
+
+    let list = global.settings.get_strv('enabled-applets');
+    for (let i = 0; i < list.length; i++) {
+        let info = list[i].split(':');
+        if (info[3] == appletDefinition.uuid && info[4] == appletDefinition.applet_id) {
+            info[0] = panelId;
+            list[i] = info.join(':');
+            break;
+        }
+    }
+
+    global.settings.set_strv('enabled-applets', list);
+}
 
 function get_role_provider(role) {
     if (Extension.Type.APPLET.roles[role]) {
@@ -387,8 +495,8 @@ function createApplet(extension, appletDefinition) {
     let orientation = appletDefinition.orientation;
     let panel_height;
 
-    panel_height = setHeightForPanel(appletDefinition.panel, appletDefinition.panel.panelPosition)
-    
+    panel_height = setHeightForPanel(appletDefinition.panel, appletDefinition.panel.panelPosition);
+
     if (appletObj[applet_id] != undefined) {
         global.log(applet_id + ' applet already loaded');
         appletObj[applet_id].setOrientation(orientation);
@@ -398,7 +506,7 @@ function createApplet(extension, appletDefinition) {
 
         return appletObj[applet_id];
     }
-    
+
     let applet;
     try {
         applet = extension.module.main(extension.meta, orientation, panel_height, applet_id);
@@ -413,6 +521,8 @@ function createApplet(extension, appletDefinition) {
     applet.instance_id = applet_id;
     applet.panel = appletDefinition.panel;
 
+    Gettext.bindtextdomain(applet._uuid, GLib.get_home_dir() + "/.local/share/locale");
+
     applet.finalizeContextMenu();
 
     return(applet);
@@ -434,58 +544,37 @@ function _removeAppletFromPanel(uuid, applet_id) {
 }
 
 function saveAppletsPositions() {
-    let zones_strings = ["left", "center", "right"];
-    let allApplets = new Array();
-    for (var i in Main.panelManager.panels){
-        let panel = Main.panelManager.panels[i];
-        if (!panel) continue;
-        for (var j in zones_strings){
-            let zone_string = zones_strings[j];
-            let zone = panel["_"+zone_string+"Box"];
-            let children = zone.get_children();
-            for (var k in children) if (children[k]._applet) allApplets.push(children[k]._applet);
-        }
-    }
-    let applets = new Array();
-    for (var i in Main.panelManager.panels){
-        let panel = Main.panelManager.panels[i];
-        if (!panel)
+    let enabled = global.settings.get_strv('enabled-applets');
+    let newEnabled = [];
+
+    for (let i = 0; i < enabled.length; i++) {
+        let info = enabled[i].split(':');
+        let applet = appletObj[info[4]];
+        if (!applet) {
             continue;
-
-        let panel_string = "panel" + i;
-
-        for (var j in zones_strings){
-            let zone_string = zones_strings[j];
-            let zone = panel["_"+zone_string+"Box"];
-            for (var k in allApplets){
-                let applet = allApplets[k];
-                let appletZone;
-                if (applet._newPanelLocation != null)
-                    appletZone = applet._newPanelLocation;
-                else
-                    appletZone = applet._panelLocation;
-                let appletOrder;
-                if (applet._newOrder != null)
-                    appletOrder = applet._newOrder;
-                else
-                    appletOrder = applet._order;
-
-                if (appletZone == zone)
-                    applets.push(panel_string+":"+zone_string+":"+appletOrder+":"+applet._uuid+":"+applet.instance_id);
+        }
+        if (applet._newOrder !== null) {
+            if (applet._newPanelId !== null) {
+                info[0] = 'panel' + applet._newPanelId;
+                info[1] = applet._zoneString;
+                applet._newPanelId = null;
             }
+            info[2] = applet._newOrder;
+            applet._newOrder = null;
+            newEnabled.push(info.join(':'));
+        }
+        else {
+            newEnabled.push(enabled[i]);
         }
     }
-    for (var i in allApplets){
-        allApplets[i]._newPanelLocation = null;
-        allApplets[i]._newOrder = null;
-    }
-    global.settings.set_strv('enabled-applets', applets);
+
+    global.settings.set_strv('enabled-applets', newEnabled);
 }
 
 function updateAppletPanelHeights(force_recalc) {
     if(!enabledAppletDefinitions)
         return;
-    
+
     for (let applet_id in enabledAppletDefinitions.idMap) {
         if (appletObj[applet_id]) {
             let appletDefinition = enabledAppletDefinitions.idMap[applet_id];
@@ -611,9 +700,9 @@ function pasteAppletConfiguration(panelId) {
     let nextId = global.settings.get_int("next-applet-id");
 
     clipboard.forEach(function(x) {
-        let uuid = x.uuid
+        let uuid = x.uuid;
         let max = Extension.get_max_instances(uuid, Extension.Type.APPLET);
-        if (max == -1 || raw.filter(a => a.split(":")[2] == uuid).length < max) {
+        if (max == -1 || raw.filter(a => a.split(":")[3] == uuid).length < max) {
             raw.push("panel" + panelId + ":" + x.location_label + ":" + x.order + ":" + uuid + ":" + nextId);
             nextId ++;
         } else {
